/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 27/11/22.
//

#ifndef FGGL_MATH_STATS_HPP
#define FGGL_MATH_STATS_HPP

#include <cstdint>
#include <cmath>

namespace fggl::math {

	class CumulativeAverage {
		public:
			inline void add(float item) {
				m_current = ( item + (m_count * m_current) ) / (m_count + 1);
				m_count++;
			}

			inline float average() {
				return m_current;
			}

		private:
			float m_current;
			std::size_t m_count;
	};

	/**
	 * Useful Statistics class.
	 * ported to C++ from Piers' Java implementation.
	 *
	 * @tparam T
	 */
	template<typename T>
	class Statistics {
		public:
			void add(T observation) {
				m_count++;
				m_sum += observation;
				m_sumSquared += (observation * observation);
				m_min = std::min(m_min, observation);
				m_max = std::max(m_max, observation);
				m_dirty = true;
			}

			inline T average() {
				compute();
				return m_average;
			}

			inline T average() const {
				if ( m_dirty ) {
					return m_sum / m_count;
				}
				return m_average;
			}

			inline T standardDeviation() {
				compute();
				return m_standardDiv;
			}

			inline T standardDeviation() const {
				if ( !m_dirty ) {
					return m_standardDiv;
				}
				T avg = average();
				T num = m_sumSquared - (m_count * avg * avg);
				return num < 0 ? 0 : sqrt( num / (m_count-1));
			}

			inline T standardError() {
				compute();
				return m_standardDiv / sqrt(m_count);
			}

			inline T standardError() const {
				if ( !m_dirty ) {
					return m_standardDiv / sqrt(m_count);
				}
				return standardDeviation() / sqrt(m_count);
			}

			inline T squareSum () const {
				return m_sumSquared;
			}

			inline std::size_t count() const {
				return m_count;
			}

			inline T sum() const {
				return m_sum;
			}

			inline T min() const {
				return m_min;
			}

			inline T max() const {
				return m_max;
			}

		private:
			T m_average{0};
			T m_sum{0};
			T m_sumSquared{0};
			T m_standardDiv{0};
			T m_min{0};
			T m_max{0};
			std::size_t m_count{0};
			bool m_dirty{false};

			void compute() {
				if ( !m_dirty ) {
					return;
				}
				m_average = m_sum / m_count;
				T num = m_sumSquared - (m_count * m_average * m_average);
				if ( num < 0 )  {
					num = 0;
				}
				m_standardDiv = sqrt( num / (m_count-1) );
				m_dirty = false;
			}
	};

	using StatisticsF = Statistics<float>;
	using StatisticsD = Statistics<double>;

} // namespace fggl::math

#endif //FGGL_MATH_STATS_HPP
